####
# bomail.guistuff.tabnothread
#
# Tabs when configuration turns off threading, i.e.
# each email gets its own line.
####

import os
import sys
import subprocess
import curses
import shlex
import functools

import bomail.config.config as config
import bomail.config.guiconfig as guicfg

import bomail.cli.mailfile as mailfile
import bomail.cli.search as search
import bomail.cli.chstate as chstate
import bomail.cli.send as send
import bomail.cli.compose as compose

import bomail.util.addr as addr
import bomail.util.datestuff as datestuff
import bomail.util.util as util

import bomail.guistuff.display as display
import bomail.guistuff.display_fmt as display_fmt
import bomail.guistuff.gui_util as gui_util


class TabNoThread:
  def __init__(self, search_str, gui):
    self.search_str = search_str
    self.gui = gui
    self.is_loaded = False
    self.cursor_ind = 0
    self.display_ind = 0
    self.num_marked = 0
    self.num_displayed = 0

    # list of lists (filename, is_marked, display_data)
    # where display_data is either None or (y_coord, lines_to_display, attr_data)
    self.file_data = []

    # for sorting filenames and tuples
    self.sort_new = not "-sortold" in shlex.split(self.search_str)
    self.file_key = lambda a: datestuff.parse_to_utc_datestr(self.gui.mail_mgr.get(a, mailfile.DATE_L))
    self.tup_key  = lambda t: self.file_key(t[0])


  # get all the filenames matching our search_str
  def load(self):
    display.draw_loading_screen(self.gui)
    filenames = search.search_argstr(self.search_str, self.gui.mail_mgr)
    self.cursor_ind = 0
    self.display_ind = 0
    self.num_marked = 0
    self.num_displayed = 0
    self.file_data = [[f, False, None] for f in filenames]
    self.file_data.sort(key=self.tup_key, reverse=self.sort_new)
    self.is_loaded = True


  # like load, but try to keep all current settings if possible
  # all possibly-important files are: those currently in our list and those matching our search string
  def recheck(self, old_disp_info):
    if not self.is_loaded:
      return
    display.draw_loading_screen(self.gui)
    new_filenames = search.search_argstr(self.search_str, self.gui.mail_mgr)
    new_fileset = set(new_filenames)
    self.remove_files([t[0] for t in self.file_data if t[0] not in new_fileset], old_disp_info)
    # assume, worst-case, that all matching files have changed/added
    self.update_for_change(new_fileset, old_disp_info)

 
  # lazily load display data and return disp_data, is_unread, is_marked
  # where disp_data = (msg_y, lines, attr_data)
  # implements memo-ization: the data is None until requested, then is loaded
  def get_display_data(self, ind):
    if self.file_data[ind][2] is None:
      height, width = self.gui.tab_area.getmaxyx()
      lines, attr_data = display_fmt.get_msg_lines_nothread(self.gui.mail_mgr, self.file_data[ind][0], width)
      self.file_data[ind][2] = [0, lines, attr_data]
    is_unread = self.file_data[ind][0] in self.gui.unread_set
    is_marked = self.file_data[ind][1]
    return self.file_data[ind][2], is_unread, is_marked


  # get these two values
  def get_old_display_info(self):
    if not self.is_loaded:
      return (None, 0)
    old_cursor_key = None if len(self.file_data) == 0 else self.tup_key(self.file_data[self.cursor_ind])
    old_display_ind = self.display_ind
    return (old_cursor_key, old_display_ind)

  # re-set display and cursor indices and display data
  def update_display(self, disp_info):
    old_cursor_key, old_display_ind = disp_info
    self.cursor_ind = 0
    self.display_ind = old_display_ind
    if old_cursor_key is not None:
      self.cursor_ind = util.bisect_left_key(self.file_data, old_cursor_key, key=self.tup_key, reverse=self.sort_new)
    display.redraw_tab_msg(self, self.gui, refresh_display=False)  # set msg_ys
    self.scroll_up(amt=0)    # if cursor and display ind are too far apart, fix them
    self.scroll_down(amt=0)


  # update this view based on the fact that these new files have been added
  # to the database
  def update_for_add(self, add_list, old_disp_info):
    if not self.is_loaded or len(add_list) == 0:
      return
    add_match_list = search.filter_argstr(self.search_str, self.gui.mail_mgr, add_list)
    if len(add_match_list) == 0:
      return
    add_tup_list = [[f, False, None] for f in add_match_list]
    add_tup_list.sort(key=self.tup_key, reverse=self.sort_new)
    self.file_data = util.merge_lists(self.file_data, add_tup_list, key=self.tup_key, reverse=self.sort_new)
    # num_marked does not need to change
    self.update_display(old_disp_info)


  # make it so that none of these files are in our view any more
  # do NOT ask mail_mgr for data of rm_list; they may not exist anymore!
  def remove_files(self, rm_list, old_disp_info):
    if not self.is_loaded or len(rm_list) == 0:
      return
    rm_set = set(rm_list)
    self.file_data = [t for t in self.file_data if t[0] not in rm_set]
    self.num_marked = sum([1 if t[1] else 0 for t in self.file_data])
    self.update_display(old_disp_info)


  # these files were deleted
  def update_for_trash(self, trash_list, data_list, old_disp_info):
    self.remove_files(trash_list, old_disp_info)


  # update based on the fact that these files have changed, or possibly
  # have been added
  def update_for_change(self, ch_list, old_disp_info):
    if not self.is_loaded or len(ch_list) == 0:
      return
    ch_set = set(ch_list)
    ch_match_set = set(search.filter_argstr(self.search_str, self.gui.mail_mgr, ch_list))
    new_file_data = []
    changed_present_set = set()
    for tup in self.file_data:
      if tup[0] in ch_match_set:
        # present and matched: reset display data
        new_file_data.append([tup[0], tup[1], None])
        changed_present_set.add(tup[0])
      elif tup[0] not in ch_set:
        # present but not changed: keep
        new_file_data.append(tup)
      # else, present, but not matched: remove
    self.file_data = new_file_data
    # add all matched, but not previously present
    add_file_list = list(ch_match_set.difference(changed_present_set))
    add_tup_list = [[f, False, None] for f in add_file_list]
    add_tup_list.sort(key=self.tup_key, reverse=self.sort_new)
    self.file_data = util.merge_lists(self.file_data, add_tup_list, key=self.tup_key, reverse=self.sort_new)
    self.num_marked = sum([1 if t[1] else 0 for t in self.file_data])
    self.update_display(old_disp_info)


  # return all filenames to be affected by the current action
  def get_curr_filenames(self):
    if not self.is_loaded:
      self.load()
    if len(self.file_data) == 0:
      return []
    elif self.num_marked == 0:
      return [self.file_data[self.cursor_ind][0]]
    else:
      return [t[0] for t in self.file_data if t[1]]  # marked


  def scroll_up(self, amt=1):
    self.cursor_ind -= amt
    if self.cursor_ind < 0:
      self.cursor_ind = 0
    if self.display_ind > self.cursor_ind:
      self.display_ind = self.cursor_ind

  def scroll_down(self, amt=1):
    self.cursor_ind += amt
    if self.cursor_ind >= len(self.file_data):
      self.cursor_ind = len(self.file_data) - 1
    if self.cursor_ind >= self.display_ind + self.num_displayed:
      self.display_ind = self.cursor_ind - int(self.num_displayed / 2)

  def scroll_down_page(self):
    self.scroll_down(self.display_ind + self.num_displayed - self.cursor_ind)

  # attempt to scroll up until the previous cursor message leaves the screen
  def scroll_up_page(self):
    old_cursor_ind = self.cursor_ind
    while self.cursor_ind > 0:
      self.cursor_ind -= 1
      display.redraw_tab_msg(self, self.gui, refresh_display=False)  # set msg_ys
      if self.cursor_ind + self.num_displayed <= old_cursor_ind:
        break
    self.display_ind = self.cursor_ind


  # return mode, note
  def process_key(self, key):
    if not self.is_loaded:
      self.load()
    mode, note = "all", ""
    curr_filelist = self.get_curr_filenames()
    
    # if no files, only allowable action is new draft
    if len(curr_filelist) == 0:
      if key == guicfg.WRITE_KEY:
        return gui_util.go_compose_draft(self.gui, "n", None)  # new blank draft
      return "note", "Key not recognized / nothing to do"

    # Now assume len(curr_filelist) > 0

    # changing state
    if key == guicfg.OPEN_KEY:
      mode, note = self.gui.acts.do(("open", curr_filelist, [self.gui.mail_mgr.get(f, mailfile.STATE_L) for f in curr_filelist]))
    elif key == guicfg.CLOSE_KEY:
      mode, note = self.gui.acts.do(("closed", curr_filelist, [self.gui.mail_mgr.get(f, mailfile.STATE_L) for f in curr_filelist]))
    elif key == guicfg.SCHEDULE_KEY:
      mode, note = gui_util.go_schedule(self.gui, curr_filelist)
    elif key == guicfg.TRASH_KEY:
      mode, note = self.gui.acts.do(("trash", curr_filelist, [self.gui.mail_mgr.get_all(f) for f in curr_filelist]))

    # tags
    elif key == guicfg.ADD_TAGS_KEY:
      mode, note = gui_util.go_add_tags(self.gui, curr_filelist)
    elif key == guicfg.REMOVE_TAGS_KEY:
      mode, note = gui_util.go_remove_tags(self.gui, curr_filelist)

    # navgation type keys
    elif key == guicfg.DOWN_KEY:
      self.scroll_down()
      mode, note = "tab", ""
    elif key == guicfg.UP_KEY:
      self.scroll_up()
      mode, note = "tab", ""
    elif key == guicfg.WAY_DOWN_KEY:
      self.scroll_down_page()
      mode, note = "tab", ""
    elif key == guicfg.WAY_UP_KEY:
      self.scroll_up_page()
      mode, note = "tab", ""

    elif (guicfg.arrowkey_nav and key == "KEY_RIGHT") or (not guicfg.arrowkey_nav and key == guicfg.RIGHT_KEY):
      # TODO: view mode
      mode, note = "tab", ""

    # mark/unmark messages
    elif key == guicfg.MARK_KEY:
      name = display_fmt.get_shortened(self.file_data[self.cursor_ind][0], self.gui.mail_mgr)
      if self.file_data[self.cursor_ind][1]:  # is marked
        self.file_data[self.cursor_ind][1] = False
        self.num_marked -= 1
        mode, note = "tab", "Un-marked " + name
      else:
        self.file_data[self.cursor_ind][1] = True
        self.num_marked += 1
      mode, note = "tab", "Marked " + name

    elif key == guicfg.MARK_ALL_KEY:
      if self.num_marked == len(self.file_data):  # unmark all
        for i in range(len(self.file_data)):
          self.file_data[i][1] = False
        self.num_marked = 0
        mode, note = "tab", "Marked none"
      else:
        for i in range(len(self.file_data)):
          self.file_data[i][1] = True
        self.num_marked = len(self.file_data)
        mode, note = "tab", "Marked all"

    # writing and viewing
    elif key == guicfg.WRITE_KEY:
      filename = curr_filelist[0]
      mode, note = gui_util.go_write_key(self.gui, filename)

    elif key == guicfg.VIEW_KEY:
      self.gui.mark_read([self.file_data[self.cursor_ind][0]])
      subprocess.call(guicfg.read_prog + " '" + self.file_data[self.cursor_ind][0] + "'", shell=True)
      curses.curs_set(False)
      return "all", ""

    # sending
    elif key == guicfg.SEND_KEY:
      filename = curr_filelist[0]
      if filename[-5:] == "draft":
        return gui_util.go_send(self.gui, filename)
      else:
        return "note", "Cannot send: not a draft"
    else:
      return "note", "Input key not recognized"
    return mode, note





####
# bomail.util.tags
#
# Each message can have a list of descriptive "tags"
# (see cli/mailfile.py).
# A list of tags is stored in config.tags_file.
# This file:
#  - utilities for dealing with the "tags" line of mailfiles
#  - the "Tag Manager" object that interacts with the tags file
####

import sys
import bisect
import os

import bomail.config.config as config
import bomail.util.merge_lines as merge_lines
import bomail.util.remove_lines as remove_lines


def clean_tag(t):
  return t.strip().replace(" ","-").replace(",","-")


def get_str_from_clean_taglist(taglist):
  return ", ".join(taglist)


def get_tagset(unclean_taglist, include_folders=False):
  new = set()
  if not include_folders:
    redundant = set()
  for t in unclean_taglist:
    w = clean_tag(t)
    if len(w) > 0:
      new.add(w)
      for i,s in enumerate(w):
        if s == "/":
          if include_folders:
            new.add(w[:i])
          else:
            redundant.add(w[:i])
  if not include_folders:
    for w in redundant:
      new.discard(w)
  return new


def get_taglist_from_list(unclean_taglist):
  return sorted(list(get_tagset(unclean_taglist, include_folders=False)))


def get_tagset_from_str(unclean_taglist_str, include_folders=False):
  return get_tagset(unclean_taglist_str.split(", "), include_folders=include_folders)


def get_taglist_from_str(unclean_taglist_str, include_folders=False):
  return sorted(list(get_tagset_from_str(unclean_taglist_str, include_folders=include_folders)))


def get_str_from_taglist(taglist):
  return get_str_from_clean_taglist(get_taglist_from_list(taglist))


# return True if search_str is equal to tag or
# is a "folder containing tag", i.e. a prefix of tag
# preceding a forward-slash
def matches(tag, search_str):
  if not tag.startswith(search_str):
    return False
  return len(tag) == len(search_str) or tag[len(search_str)] == "/"


# get all tags from taglist that match searchlist
def get_matching(taglist, searchlist):
  # TODO: better than quadratic implementation (usually doesn't matter though...)
  mylist = []
  for t in taglist:
    for s in searchlist:
      if matches(t, s):
        mylist.append(t)
  return mylist


def get_nonmatching(taglist, searchlist):
  res = []
  for t in taglist:
    if len(searchlist) == 0 or not any([matches(t,s) for s in searchlist]):
      res.append(t)
  return res


# remove all folder-prefixes from list
def remove_redundant(taglist):
  res = []
  for i,t in enumerate(taglist):
    if i < len(taglist)-1 and matches(taglist[i+1], t):
      continue
    else:
      res.append(t)
  return res


# get all tags from both lists
# but remove redundancies created by folders
def join(taglist_1, taglist_2):
  longlist = sorted(taglist_1 + taglist_2)
  return remove_redundant(longlist)


# to read from this, just use tag_mgr.tags
class TagMgr:
  def __init__(self):
    self.tags = []
    if not os.path.exists(config.tags_file):
      with open(config.tags_file, "w") as f:
        pass
    else:
      with open(config.tags_file) as f:
        for line in f.readlines():
          w = clean_tag(line)
          if len(w) > 0:
            self.tags.append(w)

  def check_and_add_tags(self, taglist):
    taglist = [t for t in taglist if t not in self.tags]
    if len(taglist) > 0:
      self.add_tags(taglist)

  def add_tags(self, taglist):
    tagset = get_tagset(taglist, include_folders=True)
    if len(tagset) > 0:
      self.tags = sorted(list(tagset.union(set(self.tags))))
      with open(config.tags_file, "w") as f:
        f.write("\n".join(self.tags))

  def remove_tags(self, taglist):
    tagset = set(taglist)
    self.tags = [t for t in self.tags if t not in tagset]
    remove_lines.do_sorted(config.tags_file, sorted(list(tagset)))

  # remove all tags except these
  def reset_tags_to(self, taglist):
    self.tags = []
    self.check_and_add_tags(taglist)





####
# bomail.util.handle
#
# Handle new emails; tag them, etc.
# Can be invoked from command line to process
# a list of already-created files.
####

import sys
import shlex

import bomail.config.config as config
import bomail.cli.chstate as chstate
import bomail.cli.mailfile as mailfile
import bomail.cli.search as search
import bomail.util.tags as tags
import bomail.util.search_opts as search_opts
import bomail.util.addr as addr
import bomail.util.tags as tags
import bomail.util.datestuff as datestuff


# A particular rule for handling new email,
# as parsed from the mail_handlers.txt file
class Rule:
  # given the text lines from the handler file
  # and the tag manager object
  def __init__(self, my_lines, tag_mgr):
    self.filter_strs = []  # filter lines from handler file
    self.querylist = []    # search query objects
    self.close = False     # close the email
    self.schedstr = None   # schedule it
    self.tags = None       # apply these tags
    self.load(my_lines, tag_mgr)

  def load(self, my_lines, tag_mgr):
    self.filter_strs = [my_lines[0]]
    ind = 1
    while ind < len(my_lines) and my_lines[ind].startswith("OR "):
      self.filter_strs.append(my_lines[ind][len("OR "):].strip())
      ind += 1

    for fs in self.filter_strs:
      self.querylist.append(search_opts.SearchQuery(liststep=False) )
      self.querylist[-1].parse(shlex.split(fs))

    act_lines = my_lines[ind:]
    self.close = False
    self.schedstr = None
    self.tags = None
    for line in act_lines:
      s = line.strip()
      if s == "close":
        self.close = True
      elif s.startswith("schedule "):
        self.schedstr = s[len("schedule "):]
      elif s.startswith("tag "):
        tagstr = s[len("tag "):]
        taglist = [tags.clean_tag(t) for t in tagstr.split(", ")]
        self.tags = taglist if self.tags is None else self.tags + taglist
    if self.tags is not None:
      self.tags.sort()
      tag_mgr.check_and_add_tags(self.tags)
      # Remove duplicates?


  # handle this list of filenames with associated datas
  # and tagsets
  # return set of indices matched by this query
  def handle(self, filelist, datalist, tagsets, mail_mgr):
    matched_indset = set()
    for q in self.querylist:
      matched_indset.update(q.filter(mail_mgr, filelist))
    if len(matched_indset) == 0:
      return matched_indset
    matched_list = [filelist[i] for i in matched_indset]

    if self.close:
      chstate.make_closed(matched_list, mail_mgr)
      mail_mgr.updated_filelist(matched_list)

    if self.schedstr is not None:
      # note: wait until now to turn schedstr into datetime because
      # it may be relative to the current time!
      chstate.schedule(matched_list, datestuff.parse_schedstr_to_utcobj(self.schedstr), mail_mgr)
      mail_mgr.updated_filelist(matched_list)

    if self.tags is not None:
      for ind in matched_indset:
        if tagsets[ind] is None:
          tagsets[ind] = set()
        tagsets[ind].update(self.tags)

    return matched_indset


# The object that takes care of handling new mail
class MailHandler:
  def __init__(self, tag_mgr=None):
    self.tag_mgr = tags.TagMgr() if tag_mgr is None else tag_mgr
    self.lastlinetags = False
    self.autotagreplies = False
    self.close_sent = False
    self.handlers = []
    self.parse_handlers()

  # read mail_handlers text file options and
  # parse into Rule objects
  def parse_handlers(self):
    try:
      with open(config.handlers_file) as f:
        lines = f.readlines()
    except:
      with open(config.error_log_file, "a") as f:
        f.write("Could not find mail-handlers file (" + config.handlers_file + ")\n\n")
      return
    ind = 0
    while ind < len(lines):
      line = lines[ind].strip()
      ind += 1
      if line == "" or line[0] == "#":
        continue

      # special options
      if line.startswith("-autotagreplies"):
        self.autotagreplies = True
        continue
      elif line.startswith("-lastlinetags"):
        self.lastlinetags = True
        continue
      elif line.startswith("-close-sent"):
        self.close_sent = True
        continue

      # asssume this is the start of a rule;
      # get all its lines until a line break or comment
      # in handler_lines
      handler_lines = [line]
      while ind < len(lines):
        line = lines[ind].strip()
        ind += 1
        if line == "" or line[0] == "#":
          break
        handler_lines.append(line)

      # create the rule
      if len(handler_lines) >= 2:
        self.handlers.append(Rule(handler_lines, self.tag_mgr))


  def handle(self, filelist, mail_mgr):
    # optimize for case with lots of tag updates
    # by not writing to file every time
    tagsets = [None]*len(filelist)
    datalist = [mail_mgr.get_all(f) for f in filelist]

    # 1. Special option: lastlinetags
    if self.lastlinetags:
      for i,data in enumerate(datalist):
        lines = data[mailfile.BODY_L].split("\n")
        if len(lines) >= 2:
          line = data[mailfile.BODY_L].split("\n")[-2]  # -1 only gave a newline/blank for some reason
          if line.startswith("tags: "):
            newtags = [tags.clean_tag(t) for t in line[len("tags: "):].split(",")]
            if tagsets[i] is None:
              tagsets[i] = set(newtags)
            else:
              tagsets[i].update(newtags)

    # 2. User-defined handlers
    for h in self.handlers:
      h.handle(filelist, datalist, tagsets, mail_mgr)

    # 3. Special option: autotagreplies
    # The problem to solve is if a message and its parent are both new
    if self.autotagreplies:
      autotag_processed = [False]*len(filelist)
      ids_to_inds = {}
      for i,d in enumerate(datalist):
        ids_to_inds[d[mailfile.MSG_ID_L]] = i
      def do_autotag(i, d):
        if autotag_processed[i]:
          return
        if tagsets[i] is None:
          tagsets[i] = set()
        autotag_processed[i] = True  # important to avoid infinite loops
        my_id = d[mailfile.MSG_ID_L]
        my_reflist = mailfile.do_get_referencelist(my_id, d[mailfile.REFS_L])
        parent_id = mailfile.do_get_parent_id(my_id, my_reflist)
        if parent_id is None:
          return
        elif parent_id in ids_to_inds:
          par_ind = ids_to_inds[parent_id]
          do_autotag(par_ind, datalist[par_ind])
          par_tags = tagsets[par_ind]
        elif parent_id in mail_mgr.ids.ids_to_filenames:
          par_tags = tags.get_tagset_from_str(mail_mgr.get(mail_mgr.ids.ids_to_filenames[parent_id], mailfile.TAGS_L))
        else:
          return
        tagsets[i].update(par_tags)
      # end of function do_autotag()

      for i,data in enumerate(datalist):
        do_autotag(i, d)

    # 4. Set tags based on handlers and options
    #    (tags get special-cased for efficiency reasons)
    for fname,tset in zip(filelist, tagsets):
      if tset is not None:
        mail_mgr.set(fname, mailfile.TAGS_L, ", ".join(sorted(tset)))

    # 5. Special option: close sent mail
    if self.close_sent:
      # close if sent and not sent to myself
      sentlist = [f for f in filelist if mail_mgr.get(f, mailfile.SENT_L) == "True" and not any([addr.is_pair_me(pr) for pr in addr.str_to_pairlist(mail_mgr.get(f, mailfile.TO_L))])]
      chstate.make_closed(sentlist, mail_mgr)



if __name__ == "__main__":
  mgr = mailfile.MailMgr()
  filenames = [s[:-1] for s in sys.stdin.readlines()]  # remove newlines
  MailHandler().handle(filenames, mgr)




####
# bomail.guistuff.display
#
# Displaying various stuff on screen.
####

import sys
import time
import curses
import curses.textpad
import subprocess
import runpy
import tempfile
import os
import re
import time
import textwrap, time

import bomail.config.config as config
import bomail.config.guiconfig as guicfg
import bomail.cli.chstate as chstate
import bomail.cli.mailfile as mailfile
import bomail.cli.compose as compose
import bomail.cli.search as search
import bomail.util.addr as addr
import bomail.util.tags as tags
import bomail.util.search_opts as search_opts


ldur_chars = guicfg.LEFT_KEY + guicfg.DOWN_KEY + guicfg.UP_KEY + guicfg.RIGHT_KEY
way_ldur_chars = guicfg.WAY_LEFT_KEY + guicfg.WAY_DOWN_KEY + guicfg.WAY_UP_KEY + guicfg.WAY_RIGHT_KEY
page_options = "  page left/down/up/right"
if guicfg.arrowkey_nav:
  ldur_chars = "←↓↑→" 
  way_ldur_chars = "Shf↔"  # shift+arrows
  page_options = "  page left/right"


ldur_str  = ldur_chars + "  left/down/up/right"
page_str  = way_ldur_chars + page_options
quit_str  = guicfg.QUIT_KEY + "  quit"
tab_str1  = guicfg.TAB_LEFT_KEY + guicfg.TAB_RIGHT_KEY + "  go tab left/right"
tab_str2  = guicfg.MOVE_TAB_LEFT_KEY + guicfg.MOVE_TAB_RIGHT_KEY + "  shift tab left/right"
tab_str3  = guicfg.ADD_TAB_KEY + guicfg.REMOVE_TAB_KEY + "  add/remove tab"
tab_str4  = guicfg.EDIT_TAB_KEY + guicfg.SEARCH_IN_TAB_KEY + "  edit/search tab"
state_str = guicfg.OPEN_KEY + guicfg.CLOSE_KEY + guicfg.SCHEDULE_KEY + guicfg.TRASH_KEY + "  open/close/schedule/trash"
tags_str  = guicfg.ADD_TAGS_KEY + guicfg.REMOVE_TAGS_KEY + "  add/delete tags"
undo_str  = guicfg.UNDO_KEY + guicfg.REDO_KEY + "  undo/redo"
get_str   = guicfg.GET_KEY + "  get updates"
write_str = guicfg.WRITE_KEY + "  write"
send_str  = guicfg.SEND_KEY + "  send"
view_str  = guicfg.VIEW_KEY + "  view"
mark_str  = guicfg.MARK_KEY + guicfg.MARK_ALL_KEY + "  mark/all"
TAB_MSG_INSTR_STR = [
    ldur_str         + "  "        + ""    + page_str  + "    "                     + " "  + undo_str + "         " + quit_str,
    "  "  + tab_str1 + "   "       + "  "  + tab_str2  + "       "                + " "   + tab_str3 + "   "            + mark_str,
    "   " + get_str  + "         " + "   " + write_str + "                      " + " " + tab_str4 + "   "            + send_str,
    "  "  + tags_str + "     "     + ""    + state_str + "  "                     + "  " + view_str]

TOPLINES = 3
BOTTOMLINES = 7


#--------------------------------------------------------------
# drawing to screen

# insert the given string to the given window with the given attribute
def my_insstr(window, y, x, s, attr):
  height, width = window.getmaxyx()
  if y >= height:
    return
  window.insstr(y, x, s.encode("utf-8")[:width-1-x], attr)

def my_addstr(window, y, x, s, attr):
  height, width = window.getmaxyx()
  if y >= height:
    return
  window.addstr(y, x, s.encode("utf-8")[:width-1-x], attr)


# write the array of msg_lines to screen,
# then set all the attributes for each (y, x, width, attr) in attr_data
def write_lines(window, y_min, y_max, y_start, x_start, msg_lines, attr_data):
  h,w = window.getmaxyx()
  for j,line in enumerate(msg_lines):
    if y_start + j >= y_min and y_start + j < y_max:
      try:
        window.insstr(y_start+j, x_start, line.encode("utf-8")[:w-x_start])
      except:  # length problem?
        window.insstr(y_start+j, x_start, line[:w-x_start])
  for y,x,width,attr in attr_data:
    if y_start + y >= y_min and y_start + y < y_max:
      window.chgat(y_start+y, x_start+x, width, attr)


def draw_loading_screen(gui):
  height, width = gui.screen.getmaxyx()
  gui.screen.clear()
  gui.screen.refresh()  # this needs to be here, no idea why
  toptabs = "[1]"
  my_insstr(gui.screen, 0, 0, toptabs, gui.attr)
  my_insstr(gui.screen, 2, 0, "-"*curses.COLS, gui.attr)
  gui.screen.chgat(0, 0, 3, curses.A_STANDOUT | gui.attr)
  gui.screen.hline(height-BOTTOMLINES, 0, "-", width, gui.attr)
  gui.screen.hline(height-BOTTOMLINES+2, 0, "-", width, gui.attr)
  redraw_note(gui, "Loading (please wait)...")
  

# mode is all, tab, or note
def redraw(gui, mode, note=""):
  if mode == "all":
    redraw_overlay(gui)
  # note this is not an elif
  if mode in ["all", "tab"]:
    redraw_tab(gui)
  note = mod_note(gui, note)
  redraw_note(gui, note)  # refreshes screen


# top lines
def redraw_overlay(gui):
  height, width = gui.screen.getmaxyx()
  gui.screen.clear()
  gui.screen.refresh()  # this needs to be here, no idea why
  toptabs = "[1]  [2]  [3]  [4]  [5]  [6]  [7]  [8]  [9]  [0]"
  if len(gui.tabs) <= 10:
    len_tt = len(gui.tabs)*5 - 2
    toptabs = toptabs[:len_tt]
  else:
    toptabs = toptabs + "  [ ]" * (len(gui.tabs) - 10)
  my_insstr(gui.screen, 0, 0, toptabs, gui.attr)
  my_insstr(gui.screen, 1, 0, gui.tabs[gui.tab_ind].search_str, gui.attr)
  my_insstr(gui.screen, 2, 0, "-"*curses.COLS, gui.attr)
  if 5*gui.tab_ind + 3 <= width:
    gui.screen.chgat(0, 5*gui.tab_ind, 3, curses.A_STANDOUT | gui.attr)  # highlight tab number
  gui.screen.hline(height-BOTTOMLINES, 0, "-", width, gui.attr)
  gui.screen.hline(height-BOTTOMLINES+2, 0, "-", width, gui.attr)


# tab section and commands
def redraw_tab(gui):
  gui.tab_area.clear()
  gui.tab_area.refresh()
  tab = gui.tabs[gui.tab_ind]
  if not tab.is_loaded:
    tab.load()
    redraw_overlay(gui)
  if guicfg.threads_on:
    if tab.mode == "thread list":
      redraw_tab_msg(tab, gui)
    elif guicfg.thread_view == "tree":
      redraw_tab_tree(tab.file_data[tab.cursor_ind][0], gui)
    else:
      redraw_tab_lin(tab.file_data[tab.cursor_ind][0], gui)
  else:
    redraw_tab_msg(tab, gui)


# choose refresh_display=False to just draw the messages
# and get the y-coordinates without refreshing the screen yet
def redraw_tab_msg(mytab, gui, refresh_display=True):
  height,screen_width = gui.tab_area.getmaxyx()

  # draw messages, updating msg_ys and num_displayed
  i = mytab.display_ind
  mytab.num_displayed = 0
  curr_y = 0
  while curr_y < height and i < len(mytab.file_data):
    display_data, is_unread, is_marked = mytab.get_display_data(i)
    mytab.file_data[i][2][0] = curr_y  # set its y-coordinate
    mytab.num_displayed += 1
    msg_lines = display_data[1]
    orig_attr_data = display_data[2] 
    attr_data = list(orig_attr_data)
    if is_unread:
      attr_data = [(y, x, w, a | curses.A_BOLD if y==0 else a) for y,x,w,a in attr_data]
    if i == mytab.cursor_ind:
      # highlight the first line
      attr_data = [(y, x, w, a | guicfg.cursor_attr if y==0 else a) for y,x,w,a in attr_data]
    write_lines(gui.tab_area, 0, height, curr_y, 0, msg_lines, attr_data)
    if is_marked:
      attr = curses.A_BOLD | gui.attr
      if i == mytab.cursor_ind:
        attr |= guicfg.cursor_attr
      my_addstr(gui.tab_area, curr_y, screen_width - 2, "X", attr)

    curr_y += len(msg_lines)
    i += 1

  if refresh_display:
    gui.tab_area.refresh()
    draw_bottom_text(gui, TAB_MSG_INSTR_STR)


def redraw_tab_lin(thread_view, gui):
  thread_view.ensure_display_loaded()
  height, screen_width = gui.tab_area.getmaxyx()

  # can assume we have > 0 files
  subj_str = gui.mail_mgr.get(thread_view.repr_file, mailfile.SUBJ_L)[:screen_width]
  my_insstr(gui.tab_area, 0, 0, subj_str, curses.A_BOLD | guicfg.DEFAULT_CLR_PAIR)
  gui.tab_area.hline(1, 0, "-", screen_width)
  offset = 2

  curr_y = thread_view.top_y_coord + offset
  for ind,tup in enumerate(thread_view.file_data):
    if tup[2]:  # is collapsed
      msg_lines, orig_attr_data = tup[3]  # collapsed disp_data
    else:
      msg_lines, orig_attr_data = tup[4]  # expanded disp data
    attr_data = list(orig_attr_data)
    if tup[0] in gui.unread_set:  # is_unread
      attr_data = [(y, x, w, a | curses.A_BOLD if y==0 else a) for y,x,w,a in attr_data]
    if ind == thread_view.cursor_ind:
      attr_data = [(y,x,screen_width,(attr if y>0 else attr|guicfg.cursor_attr)) for y,x,screen_width,attr in attr_data]
    write_lines(gui.tab_area, offset, height, curr_y, 0, msg_lines, attr_data)
    if curr_y >= 0 and curr_y < height and tup[1]:  # is marked
      attr = curses.A_BOLD | gui.attr
      if ind == thread_view.cursor_ind:
        attr |= guicfg.cursor_attr
      gui.tab_area.addch(curr_y, screen_width - 2, "X", attr)
    if curr_y-1 >= 0 and curr_y+1 < height and tup[2]:  # collapsed
      gui.tab_area.addstr(curr_y+1, 1, "<--")

    curr_y += len(msg_lines)

  gui.tab_area.refresh()
  draw_bottom_text(gui, TAB_MSG_INSTR_STR)


# TODO - implement!
def redraw_tab_tree(thread_view, gui):
  return redraw_tab_lin(thread_view, gui)


def draw_bottom_text(gui, lines):
  height, width = gui.commands_area.getmaxyx()
  gui.commands_area.clear()
  for i,line in enumerate(lines):
    if i >= height:
      break
    my_insstr(gui.commands_area, i, 0, lines[i], gui.attr)
  gui.commands_area.refresh()


# add "message x of y" to end of note
# return new note
def mod_note(gui, note):
  mytab = gui.tabs[gui.tab_ind]
  if guicfg.threads_on:
    if mytab.mode == "thread list":
      num_msgs = len(mytab.file_data)
      ind = mytab.cursor_ind
      count_note = "No threads" if num_msgs == 0 else "Thread " + str(ind+1) + " of " + str(num_msgs)
    elif mytab.mode == "one thread":
      num_msgs = len(mytab.file_data[mytab.cursor_ind][0].all_files)
      ind = mytab.file_data[mytab.cursor_ind][0].cursor_ind
      count_note = "No messages" if num_msgs == 0 else "Message " + str(ind+1) + " of " + str(num_msgs)
  else:
    num_msgs = len(mytab.file_data)
    ind = mytab.cursor_ind
    count_note = "No messages" if num_msgs == 0 else "Message " + str(ind+1) + " of " + str(num_msgs)
  height, width = gui.screen.getmaxyx()
  gap = width - (len(note) + len(count_note) + 4)
  if gap >= 0:
    note += " "*(gap + 3) + count_note
  return note



def redraw_note(gui, note):
  gui.note_area.clear()
  my_insstr(gui.note_area, 0, 0, note, gui.attr)
  gui.note_area.refresh()
  gui.screen.refresh()




####
# bomail.cli.chstate
#
# Change state of files (open/closed/scheduled/trash).
####

import sys
import shutil
import os, subprocess
import tempfile
from subprocess import Popen, PIPE

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.util.remove_lines as remove_lines
import bomail.util.merge_lines as merge_lines
import bomail.util.util as util
import bomail.util.datestuff as datestuff

usage_str = """
Change the state of an email to open, closed, or scheduled.
Reads a list of filenames from stdin, one per line, and applies the operation.
Call with no args or -h to print this help.

  bomail chstate -o           # make filenames open
  bomail chstate -c           # make filenames closed
  bomail chstate -t           # trash filenames
  bomail chstate -ut          # untrash filenames
  bomail chstate -s datestr   # schedule filename for datestr (e.g. yyyy-mm-dd)
"""


# remove files from open and scheduled lists
def remove_from_lists(filelist):
  if os.path.exists(config.openlist_file):
    remove_lines.do_sorted(config.openlist_file, sorted(filelist))
  if os.path.exists(config.scheduledlist_file):
    remove_lines.do_sched(config.scheduledlist_file, set(filelist))


# change all files' state to open
def make_open(filelist, mgr, remove_old=True):
  if remove_old:
    remove_from_lists(filelist)
  # sort openlist by lexicographic
  merge_lines.do(config.openlist_file, sorted(filelist))
  for f in filelist:
    mgr.set(f, mailfile.STATE_L, "open")


# change all files' state to closed
def make_closed(filelist, mgr):
  remove_from_lists(filelist)
  for f in filelist:
    mgr.set(f, mailfile.STATE_L, "closed")


def scheduled_list_str(filename, sched_datestr):
  return sched_datestr + " " + filename


# schedule all files for dateobj, a datetime object
def schedule(filelist, dateobj, mgr):
  remove_from_lists(filelist)
  datestr = datestuff.get_printed_schedulestr(dateobj)
  # sort scheduledlist
  lines = [scheduled_list_str(f, datestr) for f in filelist]
  merge_lines.do(config.scheduledlist_file, lines)
  for f in filelist:
    mgr.set(f, mailfile.STATE_L, "scheduled " + datestr)


# trash all files, including remove them from msg_ids_file
# return corresponding list of new filenames (in trash folder)
def trash(filelist, mgr):
  remove_from_lists(filelist)
  if not os.path.exists(config.trash_dir):
    os.makedirs(config.trash_dir)
  newnames = []
  for f in filelist:
    d = mgr.get_all(f)
    new_dir,new_ext = mailfile.get_trashfilepath(f, d)
    new_name = new_dir + new_ext
    os.makedirs(new_dir, exist_ok=True)
    if os.path.exists(f):
      util.mv_file(f, new_name)
    else:
      mailfile.write_to_file(new_name, d)
    if len(d[mailfile.ATTACH_L].strip()) > 0:
      attach_dir = mailfile.get_attach_dir(d)
      if os.path.exists(attach_dir):
        trash_attach_dir = mailfile.get_trash_attach_dir(f, d)
        os.makedirs(trash_attach_dir, exist_ok=True)
        for attach_f in os.listdir(attach_dir):
          util.mv_file(attach_dir + attach_f, trash_attach_dir + attach_f)
    newnames.append(new_name)
  mgr.trash(filelist)
  return newnames


def do_untrash(trashed_filenames, mgr):
  datalist = [mailfile.file_to_data(f) for f in trashed_filenames]
  create_filelist = ["".join(mailfile.get_mailfilepath(d)) if f[-4:] == "mail" else "".join(mailfile.get_draftfilepath(d)) for d,f in zip(datalist, trashed_filenames)]
  untrash(create_filelist, datalist, mgr)


# given list of new filenames and data to put in them
def untrash(create_filelist, datalist, mgr):
  # 1) create mailfiles
  create(create_filelist, datalist, mgr)
  # 2) move attachments back if needed
  for fname,d in zip(create_filelist,datalist):
    if len(d[mailfile.ATTACH_L].strip()) > 0:
      trash_attach_path = mailfile.get_trash_attach_dir(fname, d)
      dest_path = mailfile.get_attach_dir(d)
      if os.path.exists(dest_path):
        shutil.rmtree(dest_path)
      util.mv_file(trash_attach_path, dest_path)
    

# the opposite of trash basically
def create(filelist, datalist, mgr):
  for fname,d in zip(filelist, datalist):
    if os.path.exists(fname):
      with open(config.error_log_file, "a") as f:
        f.write("Creating file, already exists: " + fname + "\n")
        d2 = mailfile.file_to_data(fname)
        for i in range(len(d)):
          if d2[i] != d[i]:
            f.write(">>>> old version " + mailfile.fields[i] + ":\n")
            f.write(d2[i] + "\n")
            f.write(">>>> new version:\n")
            f.write(d[i] + "\n")
        f.write("\n")
    elif d[mailfile.MSG_ID_L] in mgr.ids.ids_to_filenames:
      with open(config.error_log_file, "a") as f:
        f.write("Creating file, non-identical file with same message ID already exists: " + mgr.ids.get(d[mailfile.MSG_ID_L]) + "\n")
        f.write("New data:\n  ")
        f.write("\n  ".join(d) + "\n")
    else:
      mailfile.write_to_file(fname, d)
  mgr.create(filelist, datalist)
  make_open(filelist, mgr, False)  # False == don't try to remove old ones from list
 

def main(args, filelist, mgr):
  filelist.sort()
  if args[0] == "-o":
    make_open(filelist, mgr)
  elif args[0] == "-c":
    make_closed(filelist, mgr)
  elif args[0] == "-t":
    trash(filelist, mgr)
  elif args[0] == "-ut":
    do_untrash(filelist, mgr)  # filelist are all currently in trash
  elif args[0] == "-s":
    sched = datestuff.parse_schedstr_to_utcobj(args[1])
    if sched is None:
      print("Could not parse scheduling date string!")
      exit(0)
    schedule(filelist, sched, mgr)


def main_cli():
  if len(sys.argv) < 2 or "-h" in sys.argv:
    print(usage_str)
    exit(0)
  mgr = mailfile.MailMgr()
  # remove each trailing newline
  main(sys.argv[1:], [s[:-1] for s in sys.stdin.readlines()], mgr)


if __name__ == "__main__":
  main_cli()




####
# bomail.util.util
#
# Various utility functions.
####

import os, sys

import bomail.config.config as config

def mv_file(src, dest):
  try:
    os.replace(src, dest)
  except Exception as e:
    with open(config.error_log_file, "a") as f:
      f.write("util: could not move file\n" + str(e) + "\n\n")
    with open(src) as f_src, open(dest, "w") as f_dest:
      f_dest.write(f_src.read())
    os.remove(src)

# merge lists, assuming they are in sorted order by key()
# using a key actually makes this significantly more complex
# partly because we don't want to call it unnecessarily
def merge_lists(a, b, key, reverse=False):
  if len(a) == 0:
    return b
  if len(b) == 0:
    return a
  new_list = []
  a_ind, b_ind = 0, 0
  a_key, b_key = key(a[0]), key(b[0])

  while True:
    choose_a = a_key >= b_key if reverse else a_key <= b_key
    if choose_a:
      new_list.append(a[a_ind])
      a_ind += 1
      if a_ind >= len(a):
        break
      a_key = key(a[a_ind])
    else:
      new_list.append(b[b_ind])
      b_ind += 1
      if b_ind >= len(b):
        break
      b_key = key(b[b_ind])

  # only one of the following loops will execute
  while b_ind < len(b):
    new_list.append(b[b_ind])
    b_ind += 1
  while a_ind < len(a):
    new_list.append(a[a_ind])
    a_ind += 1
  return new_list


# Input a is a list of values, sorted according to key().
# x_key is an output of key().
# Binary search, giving the position in a to insert x_key
# in order to maintain sorted order, breaking ties to the left.
# Based on the python 3 library's bisect function.
ident = lambda x: x
def bisect_left_key(a, x_key, key=ident, reverse=False, lo=0, hi=None):
  if hi is None:
    hi = len(a)
  while lo < hi:
    mid = int((lo+hi) / 2)
    mid_key = key(a[mid])
    # if it were bisect_right, we'd use <=
    is_above = (x_key < mid_key) if reverse else (mid_key < x_key)
    if is_above:
      lo = mid+1
    else:
      hi = mid
  return lo




####
# bomail.util.msgids
#
# Given a message-id, look up if we have an email file for it.
# Basically a wrapper for reading and writing to plain text file
# config.ids_to_filenames.
####

import os, sys

import bomail.config.config as config
import bomail.util.merge_lines as merge_lines
import bomail.util.remove_lines as remove_lines


def msgid_file_line(msgid, filename):
  return msgid.strip() + " " + filename


def get_idfile_lines(msg_id_list, filename_list):
  return sorted([msgid_file_line(m, f) for m,f in zip(msg_id_list,filename_list)])


class Ids:
  def __init__(self):
    self.ids_to_filenames = {}
    if os.path.exists(config.msg_ids_file):
      with open(config.msg_ids_file) as f:
        for line in f:
          prs = line.split()
          if len(prs) != 2:
            continue
          self.ids_to_filenames[prs[0].strip()] = prs[1].strip()

  def get(self, msg_id):
    return self.ids_to_filenames[msg_id]

  def add(self, msg_id_list, filename_list):
    for m,f in zip(msg_id_list, filename_list):
      self.ids_to_filenames[m] = f
    merge_lines.do(config.msg_ids_file, get_idfile_lines(msg_id_list, filename_list))

  def remove(self, msg_id_list, filename_list):
    for m in msg_id_list:
      del self.ids_to_filenames[m]
    remove_lines.do_sorted(config.msg_ids_file, get_idfile_lines(msg_id_list, filename_list))


